/*
** Module   : RuntimeJastInterpreter.java
** Abstract : Interpreter for a JAST tree that contains a compiling unit. 
**
** Copyright (c) 2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description----------------------------------
** 001 OM  20150123 Created initial version.
** 002 OM  20150210 Added support for anonymous WhereExpression and LogicalExpression.
**                  Added support for calling methods in a super class. Refactoring method names.
** 003 OM  20150306 The strings literals are generated as ready to be written to java sources and
**                  they had to be escaped to obtain their real value.
** 004 OM  20150318 Added Date and IntegerExpression implementation.
*/
package com.goldencode.p2j.persist;

import com.goldencode.ast.*;
import com.goldencode.expr.*;
import com.goldencode.p2j.uast.*;
import com.goldencode.p2j.util.*;
import java.lang.reflect.*;
import java.util.*;

/**
 * This class interprets a JAST tree. It is designed to be as generic as possible.
 * 
 * It will use the annotations from a memory stored JAST if they are found. Otherwise it will
 * use reflection to select the appropriate method/constructor based on the actual parameters.
 * In the second case, the newly discovered piece of information will be cached back in the tree
 * for a fast evaluation in successive executions.
 * 
 * The interpreter is week-typed, it only distinct two types of data: table buffers and variable.
 * variables can store any kind of data, the scalars are stored in their respective wrappers,
 * access being done using automatic boxing/unboxing. The datatypes are evaluated at runtime based
 * on the actual variable value.
 * 
 * Known issues. At this moment:
 * <ul>
 *    <li>only sequential methods can be processed. Future versions will add support for loops
 *          and conditionals;
 *    <li>there is no [this] notion. In consequence other methods cannot be called. Not even the
 *          static ones;
 *    <li>no support for arrays;
 *    <li>no support for constructors. Objects of this class cannot be built/returned. All
 *          methods are called as they were static.
 *    <li>the root of the JAST is a COMPILE-UNIT that can hold only one (public) class;
 *    <li>others TBA.
 * </ul>
 *  
 * Usage. There are 4 steps when working with the interpreter.
 * <ol>
 *    <li>Initialization. When an object is constructed, two optional types of data can be
 *       optionally supplied: a list of buffers and a set of pre-initialized variables that will
 *       be used in the process of evaluation.
 *    <li>Preparation. The JAST is supplied. The first traversal will collect paths, class
 *       variables and method names.
 *    <li>Interpretation. A method can be called. At this moment it cannot have any parameters.
 *       As workaround, they can be provided as pre-initialized variables in Initialization.
 *       This step can be repeated with different methods. If they are functions, the evaluated
 *       values is provided in the returned value. 
 *    <li>Get the variable value. Extract the value of a variable. This can be combined with
 *       [Interpretation], and can be called multiple times. If the method invoked is a function,
 *       this step is optional, a value is obtained as teh result of the [Interpretation].
 * </ol>
 */
public class RuntimeJastInterpreter
{
   /** The list of known buffers, by their name. */
   private Map<String, Buffer> buffers = null;
   
   /** The list of the variables, by their name. */ 
   private Map<String, Object> variables = null;
   
   /** The list of import paths. Used for class and method dynamic resolution. */
   private String[] imports = {
         "java.lang.",               // always present
         "com.goldencode.p2j.util.", // usually needed
   };
   
   /** The list of statically imported classes. Used for method dynamic resolution. */
   private Class<?>[] staticImports = {
         java.lang.Math.class,   // always present
   };
   
   /** The list of the method defined in this class. */
   private Map<String, Aast> classMethods = new HashMap<>();
   
   /**
    * The default constructor. Initiate the list of buffers and variables to empty sets.
    */
   public RuntimeJastInterpreter()
   {
      this.buffers = new HashMap<>();
      this.variables = new HashMap<>();
   }
   
   /**
    * A constructor that receives a set of buffers and pre-initialized variables.
    * 
    * @param   buffers
    *          The list of buffers to be used. Must be open and ready to use.
    * @param   variables
    *          The list of variables (values mapped by their variable names).
    */
   public RuntimeJastInterpreter(List<Buffer> buffers, Map<String, Object> variables)
   {
      setBuffers(buffers);
      setVariables(variables);
   }
   
   /**
    * Updates the set of known buffers. The old list is lost.
    * 
    * @param   buffers
    *          The new list of buffers to be used. Must be open and ready to use.
    */
   public void setBuffers(List<Buffer> buffers)
   {
      // map buffers for quick access by their alias
      this.buffers = new HashMap<>();
      if (buffers != null)
      {
         for (Buffer buf : buffers)
         {
            RecordBuffer rbuf = ((BufferImpl) buf).buffer();
            this.buffers.put(rbuf.getDMOAlias(), buf);
         }
      }
   }
   
   /**
    * Updates the set of known variables. The old list is lost.
    *
    * @param   variables
    *          The new list of variables (values mapped by their variable names).
    */
   public void setVariables(Map<String, Object> variables)
   {
      this.variables = new HashMap<>();
      if (variables != null)
      {
         this.variables.putAll(variables);
      }
   }
   
   /**
    * The preparation step. The node provided must be a COMPILE-UNIT. The method does a fast first
    * tree traversal (max two level deep) and will collect paths, class variables and method names
    * that will be used in future processing.
    * 
    * @param   jast
    *          The JAST tree to be processed. It must be a COMPILE-UNIT node.
    *          
    * @throws  RuntimeJastInterpreter.InterpreterException
    *          if the argument is null or not a COMPILE-UNIT node.
    */
   public void prepare(JavaAst jast)
   {
      if (jast == null || jast.getType() != JavaTokenTypes.COMPILE_UNIT)
      {
         throw new InterpreterException(
               "The argument for prepare must be a COMPILE-UNIT jast.", jast, null);
      }
      compileUnit(jast);
   }
   
   /**
    * Obtain the value of a variable. This method can be called for each needed variable, at any
    * given moment.
    * 
    * @param   varName
    *          the name of the variable whose value will be returned.
    *
    * @return  the value of requested variable.
    *
    * @throws  RuntimeJastInterpreter.InterpreterException
    *          if the argument is null or not a variable name.
    */
   public Object getVariableValue(String varName)
   {
      if (variables.containsKey(varName))
      {
         return variables.get(varName);
      }
      
      throw new InterpreterException(
            "No variable named '" + varName + "' was defined", null, null);
   }
   
   /**
    * Executes a method. If the method returns a value (function) it it will be obtained as the
    * result.
    * Only simple methods, without parameters are supported. As workaround, the parameters can be
    * added to the list of initial pre-initialized variables.
    * 
    * @param   methodName
    *          The name of the method to be called.
    *
    * @return  if the method defines a RETURN statement, that value is returned, otherwise null.
    * 
    * @throws  RuntimeJastInterpreter.InterpreterException
    *          if the argument is null or not a method name or other exception occurred during
    *          the evaluation.
    */
   public Object interpret(String methodName)
   {
      Aast aMethod = classMethods.get(methodName);
      if (aMethod != null)
      {
         return execMethod(aMethod);
      }
      throw new InterpreterException("No such method defined", null, null);
   }
   
   /**
    * Prepares the JAST for use. The method does a fast first tree traversal (max two level deep)
    * and will collect paths, class variables and method names that will be used in future
    * processing.
    * 
    * @param   cu
    *          The main node of a tree. It must be a COMPILE-UNIT node.
    */
   private void compileUnit(Aast cu)
   {
      // iterate to 'learn' how to resolve classes and (static) method  calls
      collectImportPaths(cu);
      collectStaticImports(cu);
      
      Aast mainClass = cu.getImmediateChild(JavaTokenTypes.KW_CLASS, null);
      if (mainClass == null)
      {
         throw new InterpreterException("Main class could not be determined", cu, null);
      }
      
      // store instance vars if any (the query should be listed here, normally)
      Aast classVars = mainClass.getImmediateChild(JavaTokenTypes.CS_INSTANCE_VARS, null);
      if (classVars != null)
      {
         declareVars(classVars);
      }
      
      // collect methods (all)
      Aast allMethods = mainClass.getImmediateChild(JavaTokenTypes.CS_INSTANCE_METHODS, null);
      if (allMethods == null)
      {
         throw new InterpreterException("No methods are defined in the class", cu, null);
      }
      
      Aast aMethod = allMethods.getImmediateChild(JavaTokenTypes.METHOD_DEF, null);
      while (aMethod != null)
      {
         classMethods.put(aMethod.getText(), aMethod);
         aMethod = (Aast) aMethod.getNextSibling();
      }
   }
   
   /**
    * Collects the import paths used for symbol resolution. Iterates KW_IMPORT nodes.
    * Previous collection of paths is dropped. The default "java.lang.*" package is automatically
    * inserted.
    * 
    * @param   cu
    *          The main node of a tree. It must be a COMPILE-UNIT node.
    */
   private void collectImportPaths(Aast cu)
   {
      Aast importPath = cu.getImmediateChild(JavaTokenTypes.KW_IMPORT, null);
      imports = new String[cu.getNumImmediateChildren(JavaTokenTypes.KW_IMPORT) + 1];
      int k = 0;
      imports[k++] = "java.lang.";
      while (importPath != null)
      {
         String packageName = importPath.getText();
         imports[k++] = packageName.substring(0, packageName.length() - 1);
         
         importPath = cu.getImmediateChild(JavaTokenTypes.KW_IMPORT, importPath);
      }
   }
   
   /**
    * Collects the static import paths used for symbol resolution. Iterates STATIC_IMPORT nodes.
    * Previous collection of classes is dropped. The default "java.lang.Math" package is
    * automatically inserted at the end.
    *
    * @param   cu
    *          The main node of a tree. It must be a COMPILE-UNIT node.
    */
   private void collectStaticImports(Aast cu)
   {
      Aast staticImport = cu.getImmediateChild(JavaTokenTypes.STATIC_IMPORT, null);
      staticImports = new Class[cu.getNumImmediateChildren(JavaTokenTypes.STATIC_IMPORT) + 1];
      int k = 0;
      while (staticImport != null)
      {
         String className = staticImport.getText();
         className = className.substring(0, className.length() - 2);
         try
         {
            staticImports[k++] = Class.forName(className);
         }
         catch (ClassNotFoundException e)
         {
            throw new InterpreterException("Failed to load static import", staticImport, e);
         }
         staticImport = cu.getImmediateChild(JavaTokenTypes.STATIC_IMPORT, staticImport);
      }
      staticImports[k] = Math.class; // add as a last resort
   }
   
   /**
    * Collects the variables defined in this class. Table buffers are ignored.
    * Iterates the children of CS_INSTANCE_VARS nodes, adding them to the list of internal
    * variables. If the case the initialization is performed.
    *
    * @param   classVars
    *          The main node of a tree. It must be a CS_INSTANCE_VARS node.
    */
   private void declareVars(Aast classVars)
   {
      Aast varDef = (Aast) classVars.getFirstChild();
      while (varDef != null) 
      {
         // add it to our class variables set 
         Aast varRef = (Aast) varDef.getFirstChild();
         if (varRef != null)
         {
            if (!varRef.getText().endsWith(".Buf"))
            {
               String varName = (String) varRef.getAnnotation("name");
               variables.put(varName, null);
               doAssign(varName, (Aast) varRef.getNextSibling());
            }
            // skip buffers, they are already collected
         }
         else
         {
            // simple declaration, without assignment
            String varName = (String) varDef.getAnnotation("name");
            variables.put(varName, null);
         }
         
         varDef = (Aast) varDef.getNextSibling();
      }
   }
   
   /**
    * Executes a method from this class by interpreting the sequence of statements from its body.
    * <p> 
    * Each statement from this method definition is executed/evaluated in sequence. If the method
    * returns a value (function) it it will be obtained as the result.
    * <p>
    * Only simple methods, without parameters are supported. As workaround, the parameters can be
    * added to the list of initial pre-initialized variables.
    *
    * @param   aMethod
    *          The method node to be processed.
    *
    * @return  if the method defines a RETURN statement, that value is returned, otherwise null.
    */
   private Object execMethod(Aast aMethod)
   {
      assert (aMethod.getType() == JavaTokenTypes.METHOD_DEF);
      
      Aast methBlock = aMethod.getImmediateChild(JavaTokenTypes.BLOCK, null);
      assert (methBlock != null);
      
      // iterate all statements from the method block
      Aast statement = (Aast) methBlock.getFirstChild();
      while (statement != null)
      {
         switch (statement.getType())
         {
            case JavaTokenTypes.KW_RETURN:
               if (statement.getNumImmediateChildren() == 0)
               {
                  return null;
               }
               else
               {
                  // this a function definition; returning the result
                  Object res = evalExpression((Aast) statement.getFirstChild());
                  return res;
               }
            
            case JavaTokenTypes.METHOD_CALL:
               callMethod(statement);
               break;
            
            case JavaTokenTypes.STATIC_METHOD_CALL:
               callStaticMethod(statement);
               break;
            
            case JavaTokenTypes.ASSIGN:
               Aast varRef = (Aast) statement.getFirstChild();
               doAssign(varRef.getText(), (Aast) varRef.getNextSibling());
               break;
            
            default:
               throw new InterpreterException("Unknown statement type", statement, null);
         }
         statement = (Aast) statement.getNextSibling();
      }
      return null;
   }
   
   /**
    * Performs a variable assignment. The variable must be already defined.
    * 
    * @param   varName
    *          The name of the variable.
    * @param   assignNode
    *          The node that contain the tree for evaluation of the variable value.
    *
    * @return  in order to support chaining, the evaluated value of the variable is returned.
    */
   private Object doAssign(String varName, Aast assignNode)
   {
      assert (assignNode.getType() == JavaTokenTypes.ASSIGN);
      
      if (!variables.containsKey(varName))
      {
         throw new InterpreterException(
               "No variable named '" + varName + "' was defined", assignNode, null);
      }
      
      Object val = evalExpression(assignNode);
      variables.put(varName, val);
      return val; 
   }
   
   /**
    * Call a constructor and returns the built object.
    * <p> 
    * The method will try to use the cached annotation from the node to speed up evaluation.
    * If not available, it will use reflection to resolve it. Once computed, the Constructor
    * object will be stored for future calls. 
    * <p>
    * Before calling the constructor, all children nodes are (recursively) evaluated.
    * 
    * @param   ctorNode
    *          A CONSTRUCTOR node used to build an object. 
    * 
    * @return  the newly constructed object is returned. Not null. 
    */
   private Object callCtor(Aast ctorNode)
   {
      assert (ctorNode.getType() == JavaTokenTypes.CONSTRUCTOR);
      
      Constructor<?> ctor = null;
      Class<?> queryClass = null;
      Object[] parValues = null;
      
      // try to use the runtime annotation to speed up the interpreter
      Object cachedCtor = getCache(ctorNode, "runtime-constructor");
      if (cachedCtor instanceof Constructor)
      {
         ctor = (Constructor) cachedCtor;
      }
      else
      {
         // try to use the cached class name, at least
         Object cachedClass = getCache(ctorNode, "runtime-class");
         if (cachedClass instanceof Class)
         {
            queryClass = (Class) cachedClass;
         }
         else
         {
            // TODO: check if the classname contains the package to load it directly
            
            for (String path : imports)
            {
               // load class, prepare constructor
               String ctorClass = path + ctorNode.getText();
               try
               {
                  queryClass = Class.forName(ctorClass);
                  break; // found it
               }
               catch (ClassNotFoundException e)
               {
                  // not the right package this class, try next package
               }
            }
            if (queryClass == null)
            {
               throw new InterpreterException(
                     "Failed to load the class for the constructor", ctorNode, null);
            }
            // cache the class:
            ctorNode.putAnnotationObject("runtime-class", queryClass);
         }
         
         int parCount = ctorNode.getNumImmediateChildren();
         Class[] parTypes = new Class[parCount];
         parValues = collectParameters(ctorNode, parTypes, 0);
         
         // look for the constructor
         try
         {
            ctor = queryClass.getDeclaredConstructor(parTypes);
         }
         catch (NoSuchMethodException e)
         {
            // attempt to do it manually (because of null values most likely)
            ctor = findMatchingConstructor(queryClass, parTypes);
            if (ctor == null)
            {
               throw new InterpreterException(
                     "Failed to lookup the constructor with required signature", ctorNode, null);
            }
         }
         // cache the c'tor:
         ctorNode.putAnnotationObject("runtime-constructor", ctor);
      }
      
      if (parValues == null)
      {
         // collect parameters now if not done already
         parValues = collectParameters(ctorNode, null, 0);
      }
      
      Object res = null;
      // invoke the constructor
      try
      {
         res = ctor.newInstance(parValues);
      }
      catch (IllegalAccessException | InvocationTargetException | InstantiationException e)
      {
         throw new InterpreterException(
               "Failed to lookup the constructor with required signature", ctorNode, e);
      }
      
      return res;
   }
   
   /**
    * Collects the list of parameters to a method/staticMethod/constructor and, optionally, the
    * signature in a pre-allocated table.
    * 
    * @param   methNode
    *          The method node whose children will be inspected. It can be any type of node,
    *          but the operation only makes sense for METHOD_CALL, STATIC_METHOD_CALL and
    *          CONSTRUCTOR nodes.
    * @param   signature
    *          Optional. If not null, at return, it will be populated with the method signature.
    *          If provided (not null), it must be allocated to match the signature size.
    * @param   startAt
    *          The number of children to be ignored. In the case of standard METHOD_CALLs, the
    *          first child is the object on which the method will be applied so it must be
    *          skipped (startAt = 1). Otherwise should be 0.
    * 
    * @return  an array with the values of the arguments for this method.
    */
   private Object[] collectParameters(Aast methNode, Class[] signature, int startAt)
   {
      int parCount = methNode.getNumImmediateChildren() - startAt;
      Object[] parValues = new Object[parCount];
      
      assert (signature == null || signature.length == parCount); 
      
      // iterate constructor parameters:
      for (int k = 0; k < parCount; k++)
      {
         parValues[k] = evalExpression(methNode.getChildAt(k + startAt));
         if (signature != null)
         {
            signature[k] = (parValues[k] == null) ? null : parValues[k].getClass();
         }
      }
      return parValues;
   }
   
   /**
    * Call a normal member method. If the method returns a value (function) it it will be
    * obtained as the result.
    * <p>
    * The method will try to use the cached annotation from the node to speed up evaluation.
    * If not available, it will use reflection to resolve it. Once computed, the Method
    * object will be stored for future calls. 
    * <p>
    * Before calling the method, all children nodes are (recursively) evaluated.
    * 
    * @param   aMethod
    *          The method node to be evaluated.
    *
    * @return  the result of calling the method on fist child, possible null
    */
   private Object callMethod(Aast aMethod)
   {
      assert (aMethod.getType() == JavaTokenTypes.METHOD_CALL);
      
      Method method = null;
      Class<?> aClass = null;
      Object[] parValues = null;
      Object obj = null;
      
      // 1st: find the object on which we invoke the method
      Aast ref = (Aast) aMethod.getFirstChild();
      switch (ref.getType())
      {
         case JavaTokenTypes.REFERENCE:
            // basic invocation: this should be a variable
            String refName = ref.getText();
            if (variables.containsKey(refName))
            {
               obj = variables.get(refName);
               break;
            }
            else if (buffers.containsKey(refName))
            {
               obj = buffers.get(refName);
               
               // NOTE: in dynamic queries, : and :: return some kind of value that is neither
               //    null, neither not null. The EQ operator always returns FALSE and NE always
               //    returns TRUE, even if compared with ? (unknown value).
               
               break;
            }
            else
            {
               throw new InterpreterException(
                     "Failed to execute method on reference", ref, null);
            }
            
         default:
            // TODO: chaining not implemented
            throw new InterpreterException(
                  "Unknown type of object to execute a method", ref, null);
      }
      aClass = obj.getClass();
      aMethod.putAnnotationObject("runtime-class", aClass);
      
      // 2nd: find the method to be called
      Object cachedMethod = getCache(aMethod, "runtime-method");
      if (cachedMethod instanceof Method)
      {
         method = (Method) cachedMethod;
      }
      else
      {
         String methodName = aMethod.getText();
         int parCount = aMethod.getNumImmediateChildren() - 1;
         Class[] parTypes = new Class[parCount];
         parValues = collectParameters(aMethod, parTypes, 1); // skip the 1st child
         
         Class<?> itClass = aClass;
         // attempt to use simple reflection for all super classes
         while (itClass != null)
         {
            try
            {
               method = itClass.getDeclaredMethod(methodName, parTypes);
               break; // if we found the method, exit the while loop 
            }
            catch (NoSuchMethodException e)
            {
               // do nothing, expected exception if method not defined here
            }
            itClass = itClass.getSuperclass();
         }
         
         if (method == null)
         {
            // attempt to do it manually (because of null values most likely)
            method = findMatchingMethod(aClass, methodName, parTypes);
         }
         
         if (method != null)
         {
            aMethod.putAnnotationObject("runtime-method", method);
         }
         else
         {
            throw new InterpreterException(
                  "Failed to detect the method to be called", aMethod, null);
         }
      }
      
      if (parValues == null)
      {
         // collect parameters now if not done already
         parValues = collectParameters(aMethod, null, 1);
      }
      
      // 3rd: do the dew: invoke the method on the object
      Object res = null;
      try
      {
         res = method.invoke(obj, parValues);
      }
      catch (IllegalAccessException | InvocationTargetException | IllegalArgumentException e)
      {
         throw new InterpreterException("Failed to execute method", aMethod, e);
      }
      return res;
   }
   
   /**
    * Call a static method. If the method returns a value (function) it it will be
    * obtained as the result.
    * <p>
    * The method will try to use the cached annotation from the node to speed up evaluation.
    * If not available, it will use reflection to resolve it. Once computed, the Method 
    * object will be stored for future calls. 
    * <p>
    * Before calling the method, all children nodes are (recursively) evaluated.
    *
    * @param   staticMethod
    *          The method node to be evaluated.
    *
    * @return  the result of calling the static method, possible null 
    */
   private Object callStaticMethod(Aast staticMethod)
   {
      assert (staticMethod.getType() == JavaTokenTypes.STATIC_METHOD_CALL);
      
      Method method = null;
      Class<?> staticClass = null;
      Object[] parValues = null;
      
      // try to use the runtime annotation to speed up the interpreter
      Object cachedMethod = getCache(staticMethod, "runtime-method");
      if (cachedMethod instanceof Method)
      {
         method = (Method) cachedMethod;
      }
      else
      {
         // try to use the cached class name, at least
         Object cachedClass = getCache(staticMethod, "runtime-class");
         if (cachedClass instanceof Class)
         {
            staticClass = (Class) cachedClass;
         }
         else
         {
            // TODO: check if the classname contains the package to load it directly
            String methodName = staticMethod.getText();
            int parCount = staticMethod.getNumImmediateChildren();
            Class[] parTypes = new Class[parCount];
            parValues = collectParameters(staticMethod, parTypes, 0);
            
            int k = methodName.indexOf('.');
            if (k != -1)
            {
               // <Class>.<methodName>: extract the class name and find the right package
               String className = methodName.substring(0, k);
               methodName = methodName.substring(k + 1);
               
               for (String path : imports)
               {
                  try
                  {
                     staticClass = Class.forName(path + className);
                     break;
                  }
                  catch (ClassNotFoundException e)
                  {
                     // don't worry, try next import path
                  }
               }
               if (staticClass == null)
               {
                  throw new InterpreterException(
                        "Failed to find class for static method", staticMethod, null);
               }
               
               // cache the class object
               staticMethod.putAnnotationObject("runtime-class", staticClass);
               
               try
               {
                  method = staticClass.getDeclaredMethod(methodName, parTypes);
               }
               catch (NoSuchMethodException e)
               {
                  // attempt to do it manually (because of null values most likely)
                  method = findMatchingMethod(staticClass, methodName, parTypes);
                  if (method == null)
                  {
                     throw new InterpreterException(
                           "Failed to find static method from class", staticMethod, e);
                  }
               }
            }
            else
            {
               // we do not know the classname, we will use the static import list
               for (Class<?> aClass : staticImports)
               {
                  try
                  {
                     method = aClass.getDeclaredMethod(methodName, parTypes);
                  }
                  catch (NoSuchMethodException e)
                  {
                     // attempt to do it manually (because of null values most likely)
                     method = findMatchingMethod(aClass, methodName, parTypes);
                  }
                  
                  if (method != null)
                  {
                     staticClass = aClass;
                     staticMethod.putAnnotationObject("runtime-class", staticClass);
                     break; // yes, we found it!
                  }
                  // otherwise, no problemo, try the next class
               }
            }
            
            if (method == null)
            {
               throw new InterpreterException(
                     "Failed to find static method from class", staticMethod, null);
            }
         }
         // cache the method name for later use
         staticMethod.putAnnotationObject("runtime-method", method);
      }
      
      if (parValues == null)
      {
         // collect parameters now if not done already
         parValues = collectParameters(staticMethod, null, 0);
      }
      
      Object res = null;
      // invoke the method
      try
      {
         res = method.invoke(null, parValues);
      }
      catch (IllegalAccessException | InvocationTargetException | IllegalArgumentException e)
      {
         throw new InterpreterException(
               "Failed to execute static method", staticMethod, e);
      }
      
      return res;
   }
   
   /**
    * Evaluates an expression to a value that is returned.
    * If the node evaluates to a constant, it is cashed for fast access.
    * If it is a reference to a buffer, the list of buffers is checked. Similar for variables.
    * 
    * @param   exprNode
    *          The JAST node that contains the expression.
    *
    * @return  the value of the expression.
    */
   private Object evalExpression(Aast exprNode)
   {
      if (exprNode.isAnnotation("runtime-value"))
      {
         // this node has been already processed and decided the result is constant
         Object res = getCache(exprNode, "runtime-value");
         return res;
      }
      
      String nodeText = exprNode.getText();
      Object res = null;
      switch (exprNode.getType())
      {
         case JavaTokenTypes.CONSTRUCTOR:
            res = callCtor(exprNode);
            break;
         
         case JavaTokenTypes.METHOD_CALL:
            res = callMethod(exprNode); // keep the result (function case)
            break;
         
         case JavaTokenTypes.CAST:
            res = evalExpression((Aast) exprNode.getFirstChild());
            if (!exprNode.isAnnotation("runtime-class"))
            {
               // if not already set
               exprNode.putAnnotationObject("runtime-class", getClass(nodeText));
            }
            break;
         
         case JavaTokenTypes.STRING:
            res = javaEscape(nodeText);
            // avoid processing java escapes the next time
            exprNode.putAnnotationObject("runtime-value", res);
            break;
         
         case JavaTokenTypes.EXPRESSION:
            res = evalExpression((Aast) exprNode.getFirstChild());
            break;
         
         case JavaTokenTypes.REFERENCE_DEF:
            if ("new Object".equals(nodeText))
            {
               Aast initializer = (Aast) exprNode.getNextSibling();
               int childrenCount = initializer.getNumImmediateChildren();
               Object[] objectArray = new Object[childrenCount];
               for (int k = 0; k < childrenCount; k++)
               {
                  objectArray[k] = evalExpression(initializer.getChildAt(k));
               }
               res = objectArray;
               break;
            }
            
            throw new InterpreterException("Unknown REFERENCE_DEF type", exprNode, null);
            
         case JavaTokenTypes.REFERENCE:
            // first check if this is a buffer
            if (buffers.containsKey(nodeText))
            {
               res = buffers.get(nodeText);
               break;
            }
            // or perhaps a variable name ?
            else if (variables.containsKey(nodeText))
            {
               res = variables.get(nodeText);
               break;
            }
            
            // TODO: add general reference lookup
            throw new InterpreterException(
                  "No REFERENCE named (" + nodeText + ")", exprNode, null);
         
         case JavaTokenTypes.NUM_LITERAL:
            if (nodeText.endsWith("L"))
            {
               nodeText = nodeText.substring(0, nodeText.length() - 1);
               res = Long.parseLong(nodeText);
            } 
            else 
            {
               res = Integer.parseInt(nodeText);
            }
            exprNode.putAnnotationObject("runtime-value", res); // save constant to cache
            break;
         
         case JavaTokenTypes.BOOL_FALSE:
            res = Boolean.FALSE;
            exprNode.putAnnotationObject("runtime-value", Boolean.FALSE); // save const to cache
            break;
         
         case JavaTokenTypes.BOOL_TRUE:
            res = Boolean.TRUE;
            exprNode.putAnnotationObject("runtime-value", Boolean.TRUE); // save const to cache
            break;
         
         case JavaTokenTypes.NULL_LITERAL:
            // a null parameter. It will make it difficult to select the correct method/c'tor
            res = null;
            exprNode.putAnnotationObject("runtime-value", null); // save const to cache
            break;
         
         case JavaTokenTypes.STATIC_METHOD_CALL:
            res = callStaticMethod(exprNode);
            break;
         
         case JavaTokenTypes.TERN_IF_ELSE:
            Aast cond = (Aast) exprNode.getFirstChild();
            Object bool = evalExpression(cond);
            if (bool instanceof Boolean)
            {
               Aast first = (Aast) cond.getNextSibling();
               if ((Boolean) bool)
               {
                  res = evalExpression(first);
               }
               else
               {
                  res = evalExpression((Aast) first.getNextSibling());
               }
               break;
            }
            else 
            {
               throw new InterpreterException(
                     "Not a logical expression type in ternary IF ", exprNode, null);
            }
         
         case JavaTokenTypes.ANON_CTOR:
            // we handle only the well known cases:
            // TODO: buildLogicalExpression() and buildWhereExpression() implementations are
            //       quite similar, they could be merged in a single generic method
            if ("WhereExpression".equals(exprNode.getText()))
            {
               res = buildWhereExpression(exprNode);
               break;
            }
            else if ("LogicalExpression".equals(exprNode.getText()))
            {
               res = buildTypedExpression(exprNode, "logical");
               break;
            }
            else if ("IntegerExpression".equals(exprNode.getText()))
            {
               res = buildTypedExpression(exprNode, "integer");
               break;
            }
            else if ("DateExpression".equals(exprNode.getText()))
            {
               res = buildTypedExpression(exprNode, "date");
               break;
            }
            // TBA: typed expressions as needed: decimal / datetime / datetimetz
            else
            {
               throw new InterpreterException("Unknown anonymous constructor", exprNode, null);
            }
         
         default:
            throw new InterpreterException("Unknown expression type", exprNode, null);
      }
      
      return res;
   }
   
   /**
    * Process java escapes, and return the in-memory string.
    * 
    * @param   escapedString
    *          A string, possible using java escapes, as it would appear in a java source.
    * 
    * @return  the in-memory value of the string with the escapes sequences converted to real
    *          characters.
    */
   private String javaEscape(String escapedString)
   {
      StringBuilder sb = new StringBuilder();
      int len = escapedString.length();
      for (int k = 0; k < len; k++)
      {
         char c = escapedString.charAt(k);
         if (c == '\\' && k + 1 < len)
         {
            char next = escapedString.charAt(k + 1);
            switch (next)
            {
               case 't': c = '\t'; k++; break;
               case 'b': c = '\b'; k++; break;
               case 'n': c = '\n'; k++; break;
               case 'r': c = '\r'; k++; break;
               case 'f': c = '\f'; k++; break;
               case '\'': c = '\''; k++; break;
               case '\"': c = '\"'; k++; break;
               case '\\': c = '\\'; k++; break;
               case 'u':
                  // unicode 4 hex digits
                  c = (char) Integer.parseInt(escapedString.substring(k + 2, k + 6), 16);
                  k += 5; // skip all 5 chars
                  break;
               case '0':
               case '1':
               case '2':
               case '3':
               case '4':
               case '5':
               case '6':
               case '7':
               case '8':
               case '9':
                  // octal 3 digits character escape
                  c = (char) Integer.parseInt(escapedString.substring(k + 1, k + 4), 8);
                  
                  k += 3; // skip all 3 chars
                  break;
            }
         }
         
         sb.append(c);
      }
      return sb.toString();
   }
   
   /**
    * Build an anonymous object that extends {@link WhereExpression}. 
    * 
    * @param   whereAnonCtor
    *          The JAST node that describe the new object to be built.
    * 
    * @return  the object built
    */
   private WhereExpression buildWhereExpression(Aast whereAnonCtor)
   {
      assert (whereAnonCtor.getType() == JavaTokenTypes.ANON_CTOR);
      // debug only: System.err.println(whereAnonCtor.dumpTree(true));
      
      Aast allMethods = whereAnonCtor.getImmediateChild(JavaTokenTypes.CS_INSTANCE_METHODS, null);
      assert (allMethods != null && allMethods.getNumImmediateChildren() != 0);
      
      JavaAst evaluateMethod = null;
      JavaAst getSubstMethod = null;
      Aast meth = allMethods.getImmediateChild(JavaTokenTypes.METHOD_DEF, null);
      while (meth != null)
      {
         if ("evaluate".equals(meth.getText()) && 
             "logical".equals(meth.getAnnotation("rettype")))
         {
            // TODO: the full signature is not checked, no other overloaded method should exist
            evaluateMethod = (JavaAst) meth;
         }
         else if ("getSubstitutions".equals(meth.getText()) &&
                  "Object[]".equals(meth.getAnnotation("rettype")))
         {
            getSubstMethod = (JavaAst) meth;
         }
         meth = allMethods.getImmediateChild(JavaTokenTypes.METHOD_DEF, meth);
      }
      
      assert (evaluateMethod != null); // evaluate() is abstract so it MUST be implemented.
      return new WhereExpressionAdapter(this, evaluateMethod, getSubstMethod);
   }
   
   /**
    * Build an anonymous object that extends some king of typed expression:
    * {@link IntegerExpression}, {@link LogicalExpression}, {@link DateExpression} etc.
    *
    * @param   typedExpr
    *          The JAST node that describe the new object to be built.
    * @param   retType
    *          The return type JAST of the {@code execute} method of the expected class type.
    *          Must not be {@code null} or empty.
    *
    * @return  the object built
    */
   private Resolvable buildTypedExpression(Aast typedExpr, String retType)
   {
      assert (typedExpr.getType() == JavaTokenTypes.ANON_CTOR);
      // debug only: System.err.println(typedExpr.dumpTree(true));
      
      Aast allMethods = typedExpr.getImmediateChild(JavaTokenTypes.CS_INSTANCE_METHODS, null);
      assert (allMethods != null && allMethods.getNumImmediateChildren() != 0);
      
      JavaAst executeMethod = null;
      Aast meth = allMethods.getImmediateChild(JavaTokenTypes.METHOD_DEF, null);
      while (meth != null)
      {
         if ("execute".equals(meth.getText()) &&
             retType.equals(meth.getAnnotation("rettype")))
         {
            // TODO: the full signature is not checked, no other overloaded method should exist
            executeMethod = (JavaAst) meth;
         }
         meth = allMethods.getImmediateChild(JavaTokenTypes.METHOD_DEF, meth);
      }
      
      switch (retType)
      {
         case "logical":
            return new LogicalExpressionAdapter(this, executeMethod);
         case "integer":
            return new IntegerExpressionAdapter(this, executeMethod);
         case "date":
            return new DateExpressionAdapter(this, executeMethod);
         default:
            throw new InterpreterException("Unknown anonymous constructor", typedExpr, null);
      }
   }
   
   /**
    * Obtain the class by its name.
    * At this moment this is only used by CAST nodes.
    * 
    * @param   nodeText
    *          The class name.
    * 
    * @return  The requested Class object.
    */
   private Class<?> getClass(String nodeText)
   {
      // this is a fast out
      if ("String".equals(nodeText))
      {
         return java.lang.String.class;
      }
      throw new InterpreterException("Unknown class type: " + nodeText, null, null);
   }
   
   /**
    * Look into the list of constructors for a class for the one that matches the list of
    * argument types. Only return a valid object if there is a clean choice. In case of
    * collisions (multiple matching) null is returned as error.
    *
    * @param   aClass
    *          The class whose constructor list is checked. 
    * @param   signature
    *          An array with the java types of the arguments of the required constructor. Use null
    *          as a wildcard for unknown or null parameter type.
    *
    * @return  the matching constructor. If none is found, null is returned. null is also returned
    *          if multiple constructors are found to match the signature 
    */
   private static Constructor findMatchingConstructor(Class aClass, Class[] signature)
   {
      Constructor ret = null;
      for (Constructor ctor : aClass.getDeclaredConstructors())
      {
         Class[] required = ctor.getParameterTypes();
         if (Function.matchSignature(required, signature, -1, true))
         {
            if (ret != null)
            {
               return null;
               // error: at least two constructors found to match the supplied parameter type list
            }
            
            // do not return yet, check for sanity
            ret = ctor;
         }
      }
      return ret;
   }
   
   /**
    * Look into the list of methods of a class for the one that matches name and the list of
    * argument types. Only return a valid object if there is a clean choice. In case of
    * collisions (multiple matching) null is returned as error.
    * The process is rather costly. the worse case scenario involves a second pass to re-analyze
    * the list overloaded methods if the first one (in strict mode) provided no solution.
    * <p>
    * This method calls {@link #findMatchingMethodFromClass}, for each super class until a method
    * with requested name is found, or reaching the top of the hierarchy, in which case
    * {@code null} is returned.
    *
    * @param   aClass
    *          The class whose constructor list is checked.
    * @param   signature
    *          An array with the java types of the arguments of the required constructor. Use null
    *          as a wildcard for unknown or null parameter type.
    *
    * @return  the matching constructor. If none is found, null is returned. null is also returned
    *          if multiple constructors are found to match the signature
    */
   private Method findMatchingMethod(Class aClass, String methodName, Class[] signature)
   {
      while (aClass != null)
      {
         Method ret = findMatchingMethodFromClass(aClass, methodName, signature);
         if (ret != null)
         {
            return ret;
         }
         aClass = aClass.getSuperclass();
      }
      return null;
   }

   /**
    * Look into the list of methods of a class for the one that matches name and the list of
    * argument types. Only return a valid object if there is a clean choice. In case of
    * collisions (multiple matching) null is returned as error.
    * The process is rather costly. the worse case scenario involves a second pass to re-analyze
    * the list overloaded methods if the first one (in strict mode) provided no solution.
    * <p>
    * This method is called only from {@link #findMatchingMethod}.
    *
    * @param   aClass
    *          The class whose constructor list is checked.
    * @param   signature
    *          An array with the java types of the arguments of the required constructor. Use
    *          {@code null} as a wildcard for unknown or {@code null} parameter type.
    *
    * @return  the matching constructor. If none is found, {@code null} is returned. {@code null}
    *          is also returned if multiple constructors are found to match the signature.
    */
   private Method findMatchingMethodFromClass(Class aClass, String methodName, Class[] signature)
   {
      Method ret = null;
      Set<Method> overloaded = new HashSet<>();
      for (Method meth : aClass.getDeclaredMethods())
      {
         if (!meth.getName().equals(methodName))
         {
            continue; // the name does not match
         }
         
         overloaded.add(meth); // collect overloaded methods for a second iteration 
         Class[] required = meth.getParameterTypes();
         if (Function.matchSignature(required, signature, -1, true)) /* strict checking */
         {
            if (ret != null)
            {
               return null;
               // error: at least two methods found to match the supplied parameter type list 
            }
            
            // do not return yet, check for sanity
            ret = meth;
         }
      }
      
      if (ret == null && !overloaded.isEmpty())
      {
         // second chance, check no-strict for overloaded methods 
         for (Method meth : overloaded)
         {
            Class[] required = meth.getParameterTypes();
            if (Function.matchSignature(required, signature, -1, false)) /* no strict this time */
            {
               // TODO: this is not correct we should refine the search the same way java does
               // Ex: if the sig is an Integer and there are two methods: Long & Double
               // Solution: sort overloaded on auto-promotions and choose the first match ?
               ret = meth;
               break;
            }
         }
      }
      
      return ret;
   }
   
   /***
    * Inspects the node annotation for a runtime cached value. By convention, such annotations
    * have "runtime-" prefix.
    * 
    * The reason of existence of this method is debugging, to print the already computed 
    * runtime-annotations for a node. 
    * 
    * @param   node
    *          The JAST node to be inspected.
    * @param   key
    *          The annotation name.
    *          
    * @return  the annotation value, or null, if annotation is not present.
    */
   private Object getCache(Aast node, String key)
   {
      Object o = node.getAnnotation(key);
      // debug only
      //if (o != null)
      //{
      //   System.err.println("Found cached " + key + " for " + node.toString() + 
      //                      "(" + node.getText() + ") = " + o);
      //}
      return o;
   }
   
   /**
    * Exception thrown within the RuntimeJastInterpreter. It contains the optional JAST node that
    * caused the exceptional situation beside of normal message. 
    */
   class InterpreterException
   extends RuntimeException
   {
      /** The node that caused the exception, possible null in some case. */
      private Aast node;
      
      /**
       * The constructor.
       *
       * @param   message
       *          Message that explains the situation.
       * @param   node
       *          The node that caused the exception, possible null in some case.
       * @param   cause
       *          Another Throwable that caused this exception, possible none. 
       */
      public InterpreterException(String message, Aast node, Throwable cause)
      {
         super(message, cause);
         this.node = node;
      }
      
      /**
       * Returns a short description of this exception, the message and  node type and text that
       * caused it if the node is  available.
       *
       * @return  short description of this exception.
       */
      @Override
      public String toString()
      {
         if (node != null)
         {
            return "Interpreter exception: " + super.getMessage() +
                   " in [" + node.toString() + ": " + node.getText() + "].";
         }
         else
         {
            return "Interpreter exception: " + super.getMessage();
         }
      }
   }
   
   /**
    * Adapter for an adapter class built using an anonymous constructor of the abstract class 
    * {@link WhereExpression}. It must implement at least the {@code evaluate} method. The
    * evaluation of result of the implemented/overridden methods is delegated to an interpreter.
    * <p> 
    * This class is immutable so the instances of it are constants, once created, the internal
    * data cannot be changed.
    */
   class WhereExpressionAdapter
   extends WhereExpression
   {
      /** 
       * The {@code RuntimeJastInterpreter} that will handle interpretation of this anonymous
       * object. Must not be {@code null}.
       */
      private RuntimeJastInterpreter interpreter;
      
      /**
       * The JAST tree that describes the {@code evaluate} method to be implemented by this 
       * object. Must not be {@code null} because {@code evaluate} is {@code abstract}.
       */
      private JavaAst evaluateMethod;
      
      /**
       * The JAST tree that describes the {@code getSubstitutions} method to be implemented by 
       * this object. Can be {@code null} in which case the method is not "overridden".
       */
      private JavaAst getSubstMethod;
      
      /**
       * The constructor. Initiate all private fields.
       * @param   interpreter
       *          The {@code RuntimeJastInterpreter} that will handle interpretation of this
       *          anonymous object. Must not be {@code null}.
       * @param   evaluateMethod
       *          The JAST tree that describes the {@code evaluate} method to be implemented by
       *          this object. Must not be {@code null} because evaluate is {@code abstract}
       * @param   getSubstMethod
       *          The JAST tree that describes the {@code getSubstitutions} method to be 
       *          implemented by this object. Can be {@code null} in which case the method is not
       *          "overridden".
       */
      public WhereExpressionAdapter(RuntimeJastInterpreter interpreter,
                                    JavaAst evaluateMethod,
                                    JavaAst getSubstMethod)
      {
         this.interpreter = interpreter;
         this.evaluateMethod = evaluateMethod;
         this.getSubstMethod = getSubstMethod;
      }
      
      /**
       * Interprets the method as it would be implemented.  
       *
       * @return  the logical value returned by applying the interpreter on the JAST tree.
       */
      @Override
      public logical evaluate(BaseDataType[] args)
      {
         // TODO: pass args parameters to method to be evaluated
         //       push them to variable list and remove back when leaving method body ? 
         Object res = interpreter.execMethod(evaluateMethod);
         return (res instanceof logical) ? (logical) res : new logical(false);
      }
      
      /**
       * Interprets the method as it would be overridden.
       *
       * @return  if not overridden returns the value of super method, otherwise, interprets the
       *          method from JAST tree.
       */
      @Override
      public Object[] getSubstitutions()
      {
         if (getSubstMethod == null)
         {
            // not overridden ? call super class' method
            return super.getSubstitutions();
         }
         
         Object res = interpreter.execMethod(getSubstMethod);
         return (Object[]) res;
      }
   }
   
   /**
    * Adapter for an adapter class built using an anonymous constructor of the class 
    * {@link LogicalExpression}. Overriding JASTs should implement at least the {@code execute}
    * method. The evaluation of result of the overridden method is delegated to the interpreter.
    * <p> 
    * This class is immutable so the instances of it are constants, once created, the internal
    * data cannot be changed.
    */
   class LogicalExpressionAdapter
   extends LogicalExpression
   {
      /**
       * The {@code RuntimeJastInterpreter} that will handle interpretation of this anonymous
       * object. Must not be {@code null}.
       */
      private RuntimeJastInterpreter interpreter;
      
      /**
       * The JAST tree that describes the {@code execute} method to be implemented by this object.
       */
      private JavaAst executeMethod;
      
      /**
       * The only constructor.
       * Takes a parameters the interpreter to which the evaluation of the {@code execute} method
       * is delegated and the JAST of the implemented method.
       * 
       * @param   interpreter
       *          The {@link RuntimeJastInterpreter} responsible with interpretation of the 
       *          methods overridden by this anonymous object.
       * @param   executeMethod
       *          The JAST that describes the {@code execute} method.
       */
      public LogicalExpressionAdapter(RuntimeJastInterpreter interpreter,
                                      JavaAst executeMethod)
      {
         this.interpreter = interpreter;
         this.executeMethod = executeMethod;
      }
      
      /**
       * Evaluates the {@code execute} method. If overridden, the {@code interpreter} is 
       * responsible for interpretation of the JAST, otherwise the super method is called.
       * 
       * @return  the interpreted result of the method.
       */
      @Override
      public logical execute()
      {
         if (executeMethod == null)
         {
            // not overridden ? call super class' method
            return super.execute();
         }
         
         Object res = interpreter.execMethod(executeMethod);
         return (logical) res;
      }
   }
   
   /**
    * Adapter for an adapter class built using an anonymous constructor of the class 
    * {@link DateExpression}. Overriding JASTs should implement at least the {@code execute}
    * method. The evaluation of result of the overridden method is delegated to the interpreter.
    * <p> 
    * This class is immutable so the instances of it are constants, once created, the internal
    * data cannot be changed.
    */
   class DateExpressionAdapter
   extends DateExpression
   {
      /**
       * The {@code RuntimeJastInterpreter} that will handle interpretation of this anonymous
       * object. Must not be {@code null}.
       */
      private RuntimeJastInterpreter interpreter;
      
      /**
       * The JAST tree that describes the {@code execute} method to be implemented by this object.
       */
      private JavaAst executeMethod;
      
      /**
       * The only constructor.
       * Takes a parameters the interpreter to which the evaluation of the {@code execute} method
       * is delegated and the JAST of the implemented method.
       * 
       * @param   interpreter
       *          The {@link RuntimeJastInterpreter} responsible with interpretation of the 
       *          methods overridden by this anonymous object. 
       * @param   executeMethod
       *          The JAST that describes the {@code execute} method. 
       */
      public DateExpressionAdapter(RuntimeJastInterpreter interpreter,
                                      JavaAst executeMethod)
      {
         this.interpreter = interpreter;
         this.executeMethod = executeMethod;
      }
      
      /**
       * Evaluates the {@code execute} method. If overridden, the {@code interpreter} is 
       * responsible for interpretation of the JAST, otherwise the super method is called.
       * 
       * @return  the interpreted result of the method.
       */
      @Override
      public date execute()
      {
         if (executeMethod == null)
         {
            // not overridden ? call super class' method
            return super.execute();
         }
         
         Object res = interpreter.execMethod(executeMethod);
         return (date) res;
      }
   }
   
   /**
    * Adapter for an adapter class built using an anonymous constructor of the class 
    * {@link IntegerExpression}. Overriding JASTs should implement at least the {@code execute}
    * method. The evaluation of result of the overridden method is delegated to the interpreter.
    * <p> 
    * This class is immutable so the instances of it are constants, once created, the internal
    * data cannot be changed.
    */
   class IntegerExpressionAdapter
   extends IntegerExpression
   {
      /**
       * The {@code RuntimeJastInterpreter} that will handle interpretation of this anonymous
       * object. Must not be {@code null}.
       */
      private RuntimeJastInterpreter interpreter;
      
      /**
       * The JAST tree that describes the {@code execute} method to be implemented by this object.
       */
      private JavaAst executeMethod;
      
      /**
       * The only constructor.
       * Takes a parameters the interpreter to which the evaluation of the {@code execute} method
       * is delegated and the JAST of the implemented method.
       * 
       * @param   interpreter
       *          The {@link RuntimeJastInterpreter} responsible with interpretation of the 
       *          methods overridden by this anonymous object. 
       * @param   executeMethod
       *          The JAST that describes the {@code execute} method. 
       */
      public IntegerExpressionAdapter(RuntimeJastInterpreter interpreter,
                                      JavaAst executeMethod)
      {
         this.interpreter = interpreter;
         this.executeMethod = executeMethod;
      }
      
      /**
       * Evaluates the {@code execute} method. If overridden, the {@code interpreter} is 
       * responsible for interpretation of the JAST, otherwise the super method is called.
       * 
       * @return  the interpreted result of the method.
       */
      @Override
      public integer execute()
      {
         if (executeMethod == null)
         {
            // not overridden ? call super class' method
            return super.execute();
         }
         
         Object res = interpreter.execMethod(executeMethod);
         return (integer) res;
      }
   }
}
